/*
 * $Id: ctype.h 347264 2009-11-19 13:44:32Z kdickman $
 *
 * ctype.h -- shim header for g++ support
 *
 * Stu Grossman, March 2007
 *
 * Copyright (c) 2007, Juniper Networks, Inc.
 * All rights reserved.
 */

/*
 * The c++ environment expects that all of the ctype functions (ie: isalpha,
 * etc...) are actual functions, not macros.  Also, the character class table
 * just has the wrong type altogether, and it's name *needs* to be _ctype_
 * since libstdc++ accesses it directly.
 */

#ifndef __LIBSTDCXX_EMBEDDED_CTYPE_H__
#define __LIBSTDCXX_EMBEDDED_CTYPE_H__

#ifdef __cplusplus
extern "C" {
#endif

#if 0
#include "../../../../include/ctype.h"
#endif

#define _U      0x01
#define _L      0x02
#define _N      0x04
#define _S      0x08
#define _P      0x10
#define _C      0x20
#define _X      0x40
#define _B      0x80

#undef isalnum
#undef isalpha
#undef iscntrl
#undef isdigit
#undef isgraph
#undef islower
#undef isprint
#undef ispunct
#undef isspace
#undef isupper
#undef isxdigit

int isalnum(int __c);
int isalpha(int __c);
int iscntrl(int __c);
int isdigit(int __c);
int isgraph(int __c);
int islower(int __c);
int isprint(int __c);
int ispunct(int __c);
int isspace(int __c);
int isupper(int __c);
int isxdigit(int __c);

extern const char _ctype_[];

static inline int
tolower(int c) 
{
    if (isupper(c)) {
        return c + 0x20;
    } else {
        return c;
    }
}

static inline int
toupper(int c)
{
    if (islower(c)) {
        return c - 0x20;
    } else {
        return c;
    }
}

#ifdef __cplusplus
}
#endif

#endif /* __LIBSTDCXX_EMBEDDED_CTYPE_H__ */
